/**
   @author Shin'ichiro Nakaoka
*/

#include "App.h"
#include "AppConfig.h"
#include "PluginManager.h"
#include "ExtensionManager.h"
#include "ItemManager.h"
#include "ProjectManager.h"
#include "MenuManager.h"
#include "OptionManager.h"
#include "TimeSyncItemEngine.h"
#include "MainWindow.h"
#include "RootItem.h"
#include "FolderItem.h"
#include "View.h"
#include "MessageView.h"
#include "ItemTreeView.h"
#include "ItemPropertyView.h"
#include "SceneView.h"
#include "TimeBar.h"
#include "FileBar.h"
#include "GraphBar.h"
#include "MultiValueSeqGraphView.h"
#include "MultiSE3SeqGraphView.h"
#include "SceneItem.h"
#include "MultiValueSeqItem.h"
#include "MultiSE3SeqItem.h"
#include "MultiAffine3SeqItem.h"
#include "Vector3SeqItem.h"
#include "ProjectPathSetEditor.h"
#include "DescriptionDialog.h"
#include "Licenses.h"
#include "MovieGenerator.h"

#include <cnoid/Config>
#include <cnoid/ValueTree>
#include <QApplication>
#include <QTextCodec>
#include <QIcon>
#include <QGLFormat>
#include <boost/bind.hpp>

#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace cnoid {

    class AppImpl : public QObject
    {
    public:
        AppImpl(App* self, int& argc, char**& argv);
        ~AppImpl();

        App* self;
        QApplication* qapplication;
        int& argc;
        char**& argv;
        PluginManager* pluginManager;
        ExtensionManager* ext;
        string appName;
        string vendorName;
        DescriptionDialog* descriptionDialog;

        void initialize(const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList);
        int exec();
        bool processCommandLineOptions();
        virtual bool eventFilter(QObject* watched, QEvent* event);

        void showInformationDialog();
        void onOpenGLVSyncToggled(bool on);
    };
}

App::App(int& argc, char**& argv)
{
    impl = new AppImpl(this, argc, argv);
}


AppImpl::AppImpl(App* self, int& argc, char**& argv)
    : self(self),
      argc(argc),
      argv(argv)
{
    descriptionDialog = 0;

    if(!getenv("QT_GRAPHICSSYSTEM")){
        QApplication::setGraphicsSystem("raster"); // to obtain better rendering performance
    }
    
    qapplication = new QApplication(argc, argv);
}


void App::initialize(const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList)
{
    impl->initialize(appName, vendorName, icon, pluginPathList);
}


void AppImpl::initialize( const char* appName, const char* vendorName, const QIcon& icon, const char* pluginPathList)
{
    this->appName = appName;
    this->vendorName = vendorName;

    setlocale(LC_ALL, ""); // for gettext

    QTextCodec::setCodecForCStrings(QTextCodec::codecForLocale());

    qapplication->setApplicationName(appName);
    qapplication->setOrganizationName(vendorName);
    qapplication->setWindowIcon(icon);

    ext = new ExtensionManager("Base", false);

    AppConfig::initialize(appName, vendorName);

    // OpenGL settings
    Mapping* glConfig = AppConfig::archive()->openMapping("OpenGL");
    QGLFormat glfmt;
    glfmt.setSwapInterval(glConfig->get("vsync", glfmt.swapInterval()));
    QGLFormat::setDefaultFormat(glfmt);
    
    MainWindow::initialize(appName, ext);
    MessageView::initialize(ext);
    RootItem::initialize(ext);
    ProjectManager::initialize(ext);

    FileBar::initialize(ext);
    TimeBar::initialize(ext);
    ItemTreeView::initialize(ext);
    ItemPropertyView::initialize(ext);
    SceneView::initialize(ext);
    GraphBar::initialize(ext);
    ext->addView(new MultiValueSeqGraphView());
    ext->addView(new MultiSE3SeqGraphView());
    TimeSyncItemEngineManager::initialize();
    FolderItem::initialize(ext);
    SceneItem::initialize(ext);
    MultiValueSeqItem::initialize(ext);
    MultiSE3SeqItem::initialize(ext);
    MultiAffine3SeqItem::initialize(ext);
    Vector3SeqItem::initialize(ext);

    ProjectPathSetEditor::initialize(ext);
    
    initializeMovieGenerator(ext);

    ext->menuManager().setPath("/Help").addItem(_("About Choreonoid Framework"))
        ->sigTriggered().connect(bind(&AppImpl::showInformationDialog, this));

    // OpenGL settings
    Action* vsyncItem = ext->menuManager().setPath("/Options").addCheckItem(_("Vertical sync in OpenGL rendering"));
    vsyncItem->setChecked(glfmt.swapInterval() > 0);
    vsyncItem->sigToggled().connect(bind(&AppImpl::onOpenGLVSyncToggled, this, _1));

    pluginManager = PluginManager::instance();

    if(pluginPathList){
        pluginManager->scanPluginFilesInPathList(pluginPathList);
    }
    pluginManager->scanPluginFilesInDirectoyOfExecFile();
    pluginManager->loadAllPlugins();

    MainWindow::instance()->installEventFilter(this);
}


App::~App()
{
    if(impl){
        delete impl;
    }
}


AppImpl::~AppImpl()
{
    AppConfig::flush();
    delete pluginManager;
    delete qapplication;
}


int App::exec()
{
    return impl->exec();
}


int AppImpl::exec()
{
    MainWindow* mainWindow = MainWindow::instance();

    processCommandLineOptions();

    if(!mainWindow->isVisible()){
        mainWindow->show();
    }

    int result = qapplication->exec();

    if(pluginManager->unloadAllPlugins()){
        delete ext;
        delete mainWindow;
    }
    
    return result;
}


bool AppImpl::eventFilter(QObject* watched, QEvent* event)
{
    MainWindow* mainWindow = MainWindow::instance();
    if(watched == mainWindow){
        if(event->type() == QEvent::Close){
            mainWindow->storeWindowStateConfig();
            event->accept();
            return true;
        } 
    }
                
    return false;
}



bool AppImpl::processCommandLineOptions()
{
    if(!ext->optionManager().parseCommandLine(argc, argv)){
        //put error messages
    }

    return false;
}


void AppImpl::showInformationDialog()
{
    if(!descriptionDialog){

        descriptionDialog = new DescriptionDialog();

        descriptionDialog->setWindowTitle(_("About Choreonoid Framework"));

        descriptionDialog->setDescription(
            QString(_("Choreonoid Framework Version %1\n")).arg(CNOID_FULL_VERSION_STRING) +
            _("\n"
              "This program has been developed by Shin'ichiro Nakaoka and Choreonoid Development Team, AIST, "
              "and is distributed as a part of the Choreonoid package.\n"
              "\n") +
            LGPLtext() +
            _("The source and some binary packages of this program also include the following thirdparty "
              "libraries:\n"
              "\n"
              "* Eigen (http://eigen.tuxfamily.org/)\n"
              "* IJG JPEG Library (http://www.ijg.org/)\n"
              "* libpng (http://www.libpng.org/pub/png/libpng.html)\n"
              "* LibYAML (http://pyyaml.org/wiki/LibYAML)\n"
              "* zlib (http://zlib.net/)\n"
              "\n"
              "These libraries are used and redistributed under the terms of their licenses. Please see "
              "the corresponding directories in the source package or their official web sites to see the "
              "details of their licenses.\n"
              "\n"
              "This program also depends on a lot of other external libraries, which are linked into this "
              "program during the execution. Some binary packages of this program may include the binaries "
              "of the following libraries:\n"
              "\n"
              "* Qt (http://qt.nokia.com/)\n"
              "* OpenSceneGraph (http://www.openscenegraph.org)\n"
              "* Boost C++ Libraries (http://www.boost.org/) \n"
              "\n"
              "Please see their official web sites to see the details of their licenses.\n"
                ));
    }

    descriptionDialog->show();
}


void AppImpl::onOpenGLVSyncToggled(bool on)
{
    Mapping* glConfig = AppConfig::archive()->openMapping("OpenGL");
    glConfig->write("vsync", (on ? 1 : 0));

    // show messag dialog to prompt restart here
}
